#include "xsnmp.h"
#include <sys/ioctl.h>

/*
 *			P I N G . C
 *
 * Using the InterNet Control Message Protocol (ICMP) "ECHO" facility,
 * measure round-trip-delays and packet loss across network paths.
 *
 * Author -
 *	Mike Muuss
 *	U. S. Army Ballistic Research Laboratory
 *	December, 1983
 * Modified at Uc Berkeley
 * Record Route and verbose headers - Phil Dykstra, BRL, March 1988.
 * Multicast options (ttl, if, loop) - Steve Deering, Stanford, August 1988.
 * ttl, duplicate detection - Cliff Frost, UCB, April 1989
 * Pad pattern - Cliff Frost (from Tom Ferrin, UCSF), April 1989
 * Wait for dribbles, option decoding, pkt compare - vjs@sgi.com, May 1989
 *
 * Status -
 *	Public Domain.  Distribution Unlimited.
 *
 * Bugs -
 *	More statistics could always be gathered.
 *	This program has to run SUID to ROOT to access the ICMP socket.
 */

#include <errno.h>

#include <sys/param.h>
#include <sys/file.h>

#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>

#define Debug	(Flags & (PROG_DEBUG_PING | PROG_DEBUG))

#define	MAXPACKET	(65536-60-8)	/* max packet size */

/* MAX_DUP_CHK is the number of bits in received table, ie the */
/*      maximum number of received sequence numbers we can keep track of. */
/*      Change 128 to 8192 for complete accuracy... */

#define MAX_DUP_CHK     8 * 128
int     mx_dup_ck = MAX_DUP_CHK;
char    rcvd_tbl[ MAX_DUP_CHK / 8 ];

#define A(bit)          rcvd_tbl[ (bit>>3) ]    /* identify byte in array */
#define B(bit)          ( 1 << (bit & 0x07) )   /* identify bit in byte */
#define SET(bit)        A(bit) |= B(bit)
#define CLR(bit)        A(bit) &= (~B(bit))
#define TST(bit)        (A(bit) & B(bit)) */

u_char	*packet;
int	i;

struct sockaddr whereto;	/* Who to ping */
#define DATALEN	64-8
int datalen = DATALEN;		/* How much data */
int packlen=DATALEN+100;

static u_char outpack[MAXPACKET];

int ntransmitted = 0;		/* sequence # for outbound packets = #sent */
int ident;

struct sockaddr_in *to = (struct sockaddr_in *) &whereto;
static u_char *datap = &outpack[8+sizeof(struct timeval)];
int	Pingsocket=0;
PINGENTRY	*Pinglist=(PINGENTRY *)NULL;

initpingsocket() {
    struct protoent *proto;
    int		i;
    
    if(! Pingsocket) {
	if((proto = getprotobyname("icmp")) == (struct protoent *)NULL) {
	    fprintf(stderr, "icmp: unknown protocol\n");
	    exit(10);
	}

	if((Pingsocket = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
	    perror("ping: socket");
	    exit(5);
	}
	/* make it nonblocking */
	i = 1;
	ioctl(Pingsocket, FIONBIO, &i);
    }
}

initpings() 
{
    XtAddInput(Pingsocket, (XtPointer)XtInputReadMask
	       , (XtInputCallbackProc)x_readsocket
	       , (XtPointer)NULL);
}

int sendping(address) u_long address; {
    struct in_addr ifaddr;
    register struct icmp *icp;
    int cc;
    char outpack[sizeof(struct icmp)];
    
    bzero((char *)&whereto, sizeof(struct sockaddr) );
    to->sin_family = AF_INET;
    to->sin_addr.s_addr = address;
    
    /*	packlen = datalen + 60 + 76;	/* MAXIP + MAXICMP */
    
    if(! ident)
	ident = getpid() & 0xFFFF;
    
    icp = (struct icmp *) outpack;
    icp->icmp_type = ICMP_ECHO;
    icp->icmp_code = 0;
    icp->icmp_cksum = 0;
    icp->icmp_seq = htons((unsigned short)(ntransmitted++));
    icp->icmp_id = ident;		/* ID */
    
    CLR( ntohs(icp->icmp_seq) % mx_dup_ck );
    
    cc = datalen+8;			/* skips ICMP portion */
    
    /* Compute ICMP checksum here */
    icp->icmp_cksum = in_cksum( (u_short *)icp, cc );
    
    /* cc = sendto(s, msg, len, flags, to, tolen) */
    sendto( Pingsocket, (char *)outpack, cc, 0, &whereto
	   , sizeof(struct sockaddr) );
}

u_long readping() {
    struct ip		*ip;
    register struct icmp *icp;
    int			hlen;
    int			i, cc;
    struct sockaddr_in	from;

    if(! packet && (packet = (u_char *)malloc(packlen)) == NULL) {
	fprintf( stderr, "ping: malloc failed for %d bytes\n", packlen);
	perror("");
	exit(1);
    }

    errno = 0;
    i = sizeof(from);
    if((cc=recvfrom(Pingsocket, (char *)packet, packlen
		    , 0, &from, &i)) < 0)
	return(NULL);

    /* Check the IP header */
    ip = (struct ip *) packet;
    hlen = ip->ip_hl << 2;

    icp = (struct icmp *)(packet + hlen);
    if( icp->icmp_type == ICMP_ECHOREPLY && icp->icmp_id == ident )
	return(from.sin_addr.s_addr);
/*	return(ip->ip_src.s_addr);	/* return from addr */
    else
	return(NULL);
}

/*
 *			I N _ C K S U M
 *
 * Checksum routine for Internet Protocol family headers (C Version)
 *
 */
in_cksum(addr, len)
u_short *addr;
int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while( nleft > 1 )  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if( nleft == 1 ) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}

/*
 * Routine to create a new pingentry
 */
PINGENTRY *newpingentry(ifptr, interval) IFENTRY *ifptr; int interval; {
    register PINGENTRY	*pingent;

    if((pingent = (PINGENTRY *)malloc(sizeof(PINGENTRY)))
       == (PINGENTRY *)NULL) {
	fprintf(stderr, "Can't malloc more memory!\n");
	exit(1);
    }
    
    pingent->pen_if = ifptr;
    pingent->pen_next = (PINGENTRY *)NULL;
    pingent->pen_interval = interval;
    pingent->pen_pings = 0;
    pingent->pen_flags = 0;

    return(pingent);
}

reinsertping(newping, interval) PINGENTRY *newping; int interval; {
    register PINGENTRY	**pingent;
    register int	count;
    register PINGENTRY	*pent;

    /* Set the interval again */
    newping->pen_interval = interval;

    /* Reset the pings counter */
    newping->pen_pings = 0;

/*    if(Debug)
/*	printf("Reinserting '%s'(%d) into ping list\n"
/*	       , newping->pen_if->if_host->hl_name
/*	       , newping->pen_interval); */

    /* Is there anything in the list? */
    if(! Pinglist) {
	Pinglist = newping;
#ifdef	DEBUG
	printf("Inserting at beginning\n");
	printpinglist();
#endif
	return(TRUE);
    }

    /* Does this one go at the beginning? */
    if(Pinglist->pen_interval > newping->pen_interval) {
	pent = newping;
	pent->pen_next = Pinglist;
	Pinglist = pent;
	/* and adjust the next one */
	pent->pen_next->pen_interval -= pent->pen_interval;
#ifdef	DEBUG
	printf("Inserting just after first one\n");
	printpinglist();
#endif
	return(TRUE);
    }

    /* There's at least one entry on the list */
    for(pingent = &Pinglist , count = Pinglist->pen_interval
	; (*pingent)->pen_next
	; pingent = &(*pingent)->pen_next) {
	if(count + (*pingent)->pen_next->pen_interval > newping->pen_interval) {
	    /* goes here */
	    pent = newping;
	    pent->pen_next = (*pingent)->pen_next;
	    (*pingent)->pen_next = pent;
	    /* make this new entry fit */
	    pent->pen_interval -= count;
	    /* and adjust the next one */
	    pent->pen_next->pen_interval -= pent->pen_interval;
	    if(pent->pen_next->pen_interval < 0)
		if(Debug) {
		    printf("INSERTING '%s'(%d) just before '%s'\n"
			   , newping->pen_if->if_host->hl_name
			   , newping->pen_interval
			   , pent->pen_next->pen_if->if_host->hl_name);
		    printpinglist();
		}
	    
	    return(TRUE);
	}
	count += (*pingent)->pen_next->pen_interval;
#ifdef	DEBUG
 	printf("skipping (count=%d)\n", count);
#endif
    }

    /* Insert this one at the end of the list */
    (*pingent)->pen_next = newping;
    newping->pen_interval -= count;
#ifdef	DEBUG    
    printf("inserting at end\n");
    printpinglist();
#endif
    return(TRUE);
}

/*
 * Insert an interface into the ping list
 */
insertping(ifptr) IFENTRY *ifptr; {
    register PINGENTRY	*pent;

#ifdef	DEBUG
    printf("Inserting '%s'(%d) into ping list\n"
	   , ifptr->if_host->hl_name
	   , hostptr->hl_interval);
#endif

    pent = newpingentry(ifptr, 0);
    pent->pen_next = Pinglist;
    Pinglist = pent;
    return;
}

/*
 * Remove a ping entry from the list
 */
PINGENTRY *extractping(address) u_long address; {
    register PINGENTRY	*pingent;
    register PINGENTRY	*pent;
    register PINGENTRY	*pret;

    if(Pinglist == (PINGENTRY *)NULL) return((PINGENTRY *)NULL);

#ifdef	DEBUG
    printf("Removing '%s' from pinglist\n", iftohost(address)->hl_name);
#endif

    /* Is it at the beginning? */
    if(Pinglist->pen_if->if_addr.s_addr == address) {
	pent = Pinglist->pen_next;
	pret = Pinglist;
	pret->pen_next = (PINGENTRY *)NULL;
	Pinglist = pent;
#ifdef	DEBUG
	fprintf(stdout, "Removed an entry from the beginning\n");
	printpinglist();
#endif
	return(pret);
    }

    for(pingent = Pinglist
	; pingent->pen_next
	; pingent = pingent->pen_next)
	if(pingent->pen_next->pen_if->if_addr.s_addr == address) {
	    /* remove this one */
	    pent = pingent->pen_next;
	    pingent->pen_next = pent->pen_next;
	    pret = pent;
	    pret->pen_next = (PINGENTRY *)NULL;
#ifdef	DEBUG
	    fprintf(stdout, "Removed an entry from the middle\n");
	    printpinglist();
#endif
	    return(pret);
	}

    printf("Can't find ping entry to remove!\n");
    printpinglist();
    return((PINGENTRY *)NULL);
}

freeping(pptr) PINGENTRY *pptr; {

    free(pptr);
}

removeping(pptr) PINGENTRY *pptr; {

    extractping(pptr->pen_if->if_addr.s_addr);

    freeping(pptr);
}

removepings(ifptr) IFENTRY *ifptr; {
    register PINGENTRY	*pptr;
    register PINGENTRY	*nextp;

    for(pptr = Pinglist ; pptr ; pptr = nextp) {
	nextp = pptr->pen_next;
	if(pptr->pen_if == ifptr)
	    removeping(pptr);
    }
}

/*
 * Print the pinglist
 */
printpinglist() {
    register PINGENTRY	*pingent;

    printf("Printing Ping List:\n");
    for(pingent = Pinglist ; pingent ; pingent = pingent->pen_next)
	printpingentry(pingent);
    printf("\nFinished printing ping list\n");
}

/*
 * Print a ping entry
 */
printpingentry(pingent) PINGENTRY *pingent; {

    printf("%s(%d) "
	   , pingent->pen_if->if_host->hl_name
	   , pingent->pen_interval);
}
